function    x_out = ADMM_log(y , lambda , D , delta , varargin)

% x_out = ADMM_log(y , lambda , D , delta , varargin)
%
% The function finds the minimizer of the following quadratic regularization:
%
%               x_out = argmin_x  ||x - y||^2 + lambda * sum_i log( delta^2 + (Ai*x)^2 )
% 
% Instead of providing the function with the matrix 'A', a function handle
% "D" is used such that 
%                       D(x,1) = A*x    and     D(x,-1) = A^T*x 
%
% Other parameter listed below are optional in the sense that if they are
% not given, a default value is used:
%
% ADMM_log(... , 'max_iter' , n , ...)
% "n" is the maximum number of iterations used for the conjugate gradient 
% method. The default value is 100.
%
% ADMM_log(... , 'Rel_err_tol' , tol , ...)
% "tol" is the tolerance in relative error (relative change of the input 
% argument in the minimization steps) in the ADMM method which is a 
% termination condition. The default value is 1e-8.
%
% ADMM_log(... , 'Anim_En' , flag , ...)
% "flag" shows if the results in the intermediate iterations are plotted or
% not. The default value is 'false'. Other inputs enable the plot.
%
% ADMM_log(... , 'Orig_Sig' , x_orig , ...)
% "x_orig" is the original noiseless signal which we would like to obtain
% by this regularization method. It is only required when the animation 
% obtion is enabled.


% updating with the given inputs
[N_iter , tol , flag , x_orig]   = process_options(varargin , 'max_iter' , 100 , ...
                'Rel_err_tol' , 1e-8 , 'Anim_En' , 'false' , 'Orig_Sig' , []);


                            

% initial tests to check if the inputs are ok
if size(y , 1) == 1
    ys  = y.';
elseif size(y , 2) == 1
    ys  = y;
else
    error('!!! "y" should be a vector !!!')
    return
end

if size(x_orig , 1) == 1
    x_orig  = x_orig.';
else
    x_orig  = x_orig;
end

    
    
if isa(D , 'function_handle') ~= 1
    error('!!! The "D" should be a function handle !!!')
    return
end



% % overal cost function
% JJJ         = @(x) sum( (x-ys).^2 ) + lambda * sum( log( D(x,1).^2 + delta^2 ) );



% initialization
x_cur       = ys;   % or 0!
z_cur       = D(x_cur , 1);
u_cur       = zeros(size(z_cur));
rho         = 2 * lambda;
it_num      = 0;
re          = 1;




% iterations
if strcmpi(flag , 'false')  % no animation
    
    while (it_num < N_iter)&(re > tol)
        it_num  = it_num + 1;
        
        x_new   = L2_Prox(ys  ,  z_cur - u_cur  ,  rho  ,  D);
        re      = norm(x_cur - x_new) / norm(x_new);
        x_cur   = x_new;
        Dx      = D(x_cur , 1);
        
        z_cur   = log_Prox(Dx + u_cur , lambda/rho , delta);
        
        u_cur   = u_cur + Dx - z_cur;
        
%         % check the overal cost
%         JJJ(x_cur)
    end
    
    
    
else
    
    
    if ~isempty(x_orig)     % with animition and with knowing the oracle solution
        
        while (it_num < N_iter)&(re > tol)
            it_num  = it_num + 1;
            
            x_new   = L2_Prox(ys  ,  z_cur - u_cur  ,  rho  ,  D);
            re      = norm(x_cur - x_new) / norm(x_new);
            x_cur   = x_new;
            Dx      = D(x_cur , 1);
            
            z_cur   = log_Prox(Dx + u_cur , lambda/rho , delta);
            
            u_cur   = u_cur + Dx - z_cur;
            
%             % check the overal cost
%             JJJ(x_cur)
            
            
            plot([1:length(ys)]' , x_orig ,[1:length(ys)]' , ys , '.' , [1:length(ys)]' , x_cur , '-.', 'LineWidth', 1.5)
            grid on;
            hold on;
            SNR     = 10*log10( mean(x_orig.^2) / mean( (x_orig - x_cur).^2 ));
            title(sprintf('[#iter. = %d/%d      SNR = %f dB       Relative Error = %f]', it_num, N_iter, SNR , re));
            legend('Main', 'Noisy', 'Est.');
            axis([-0.5 , length(ys)+0.5 , min(ys)-1 , max(ys)+1])
            hold off;
            getframe;
            
        end
        

        
    else                    % with animition but without knowing the oracle solution
        
        while (it_num < N_iter)&(re > tol)
            it_num  = it_num + 1;
            
            x_new   = L2_Prox(ys  ,  z_cur - u_cur  ,  rho  ,  D);
            re      = norm(x_cur - x_new) / norm(x_new);
            x_cur   = x_new;
            Dx      = D(x_cur , 1);
            
            z_cur   = log_Prox(Dx + u_cur , lambda/rho , delta);
            
            u_cur   = u_cur + Dx - z_cur;
            
%             % check the overal cost
%             JJJ(x_cur)
            
            plot([1:length(ys)]' , ys , '.' , [1:length(ys)]' , x_cur , '-.', 'LineWidth', 1.5)
            grid on;
            hold on;
            title(sprintf('[#iter. = %d/%d      Relative Error = %f]', it_num, N_iter , re));
            legend('Noisy', 'Est.');
            axis([-0.5 , length(ys)+0.5 , min(ys)-1 , max(ys)+1])
            hold off;
            getframe;
            
        end
        
    end
end


% output
x_out       = reshape(x_cur , size(y));




%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Auxiliary functions called within this function
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    %--------------------%
    %----------------------------------------%
    %------------------------------------------------------------%
    function x_min = L2_Prox(y , t , lambda , D)
        % The function finds the minimizer of
        %       J(x) = ||x-y||^2 + lambda * ||Ax-t||^2
        % "D" is a function handle such that
        %       D(x,1) = A*x    and     D(x,-1) = A^T*x 
        
%         % cost function
%         JJ  = @(x) sum( (x-y).^2 ) + lambda * sum( (D(x,1) - t).^2 );
        
        % initialization
        NN          = 100;
        TOL         = 1e-4;
        y_mdf       = y + lambda * D(t , -1);
        x_min       = y_mdf;   % or 0!
        r_old       = y_mdf - IpDTD(lambda , x_min);
        r_old       = y_mdf - lambda * D(D(x_min , 1) , -1) - x_min;
        nrm_ro      = sum(r_old.^2);
        p           = r_old;
        itr         = 0;
        
        % CG iterations
        while (itr < NN)&(nrm_ro > TOL)
            itr     = itr + 1;
            
            Ap      = lambda * D(D(p , 1) , -1) + p;
            alpha   = nrm_ro / (p' * Ap);
            
            x_min   = x_min + alpha * p;
            
            r_new   = r_old - alpha * Ap;
            nrm_rn  = sum(r_new.^2);
            
            p       = r_new + nrm_rn/nrm_ro * p;
            
            r_old   = r_new;
            nrm_ro  = nrm_rn;
            
            %JJ(x_min)
            
        end
    end
    %------------------------------------------------------------%
    %----------------------------------------%
    %--------------------%
    
    
    %--------------------%
    %----------------------------------------%
    %------------------------------------------------------------%
    function z_min = log_Prox(y , lambda , delta)
        % The function finds the minimizer of
        %       J(z) = ||z-y||^2 + lambda * sum_i log(z_i^2 + delta^2)
        % Note that the cost is separable
        
        shC     = y/3;
        
        p3      = (lambda + delta^2) / 3 - shC.^2;
        q2      = - shC .* (shC.^2 - lambda/2 + delta^2);
        
        DD      = p3.^3 + q2.^2;
        
        ind     = (DD > 0);
        indC    = logical(1 - ind);
        
        x_min   = zeros(size(y));
        
        if sum(ind) > 0
            sDD = sqrt(DD(ind));
            z_min(ind)  = shC(ind) + sign( -q2(ind) + sDD ) .* abs( -q2(ind) + sDD ).^(1/3) ...
                                   + sign( -q2(ind) - sDD ) .* abs( -q2(ind) - sDD ).^(1/3);
        end
        
        if sum(indC) > 0
            theta   = acos(-q2(indC) ./ (-p3(indC)) .^ 1.5);
            ampl    = 2 * sqrt( -p3(indC) );
            
            z1      = shC(indC) + ampl .* cos(theta / 3);
            z2      = shC(indC) + ampl .* cos((theta + 2*pi) / 3);
            z3      = shC(indC) + ampl .* cos((theta - 2*pi) / 3);
            
            y1      = (z1 - y(indC)) .^ 2 + lambda * log(z1.^2 + delta^2);
            y2      = (z2 - y(indC)) .^ 2 + lambda * log(z2.^2 + delta^2);
            y3      = (z3 - y(indC)) .^ 2 + lambda * log(z3.^2 + delta^2);
            
            [val inD]   = min([reshape(y1 , 1 , length(y1)) ; reshape(y2 , 1 , length(y2)) ; reshape(y3 , 1 , length(y3))]);
            inD     = reshape(inD , size(z1));
            z_min(indC) = z1 .* (inD == 1) + z2 .* (inD == 2) + z3 .* (inD == 3);
        end
        
        z_min       = reshape(z_min , size(y));
    end
    %------------------------------------------------------------%
    %----------------------------------------%
    %--------------------%
    
    
    %--------------------%
    %----------------------------------------%
    %------------------------------------------------------------%
    function w = IpDTD(lambda , v);
        % the function implements   w =  (I + lambda* D^T*D) * v
        
        w       = v;
        for col = 1 : size(v , 2)
            w(: , col)  = w(: , col) + lambda * D( D(v(: , col) , 1) , -1);
        end
    end
    %------------------------------------------------------------%
    %----------------------------------------%
    %--------------------%



    %--------------------%
    %----------------------------------------%
    %------------------------------------------------------------%
    % PROCESS_OPTIONS - Processes options passed to a Matlab function.
    %                   This function provides a simple means of
    %                   parsing attribute-value options.  Each option is
    %                   named by a unique string and is given a default
    %                   value.
    % Copyright (C) 2002 Mark A. Paskin
    function [varargout] = process_options(args, varargin)
        
        % Check the number of input arguments
        n = length(varargin);
        if (mod(n, 2))
            error('Each option must be a string/value pair.');
        end
        
        % Check the number of supplied output arguments
        if (nargout < (n / 2))
            error('Insufficient number of output arguments given');
        elseif (nargout == (n / 2))
            warn = 1;
            nout = n / 2;
        else
            warn = 0;
            nout = n / 2 + 1;
        end
        
        % Set outputs to be defaults
        varargout = cell(1, nout);
        for i=2:2:n
            varargout{i/2} = varargin{i};
        end
        
        % Now process all arguments
        nunused = 0;
        for i=1:2:length(args)
            found = 0;
            for j=1:2:n
                if strcmpi(args{i}, varargin{j})
                    varargout{(j + 1)/2} = args{i + 1};
                    found = 1;
                    break;
                end
            end
            if (~found)
                if (warn)
                    warning(sprintf('Option ''%s'' not used.', args{i}));
                    args{i}
                else
                    nunused = nunused + 1;
                    unused{2 * nunused - 1} = args{i};
                    if length(args)>i
                        unused{2 * nunused} = args{i + 1};
                    end
                end
            end
        end
        
        % Assign the unused arguments
        if (~warn)
            if (nunused)
                varargout{nout} = unused;
            else
                varargout{nout} = cell(0);
            end
        end
    end
    %------------------------------------------------------------%
    %----------------------------------------%
    %--------------------%





    
end