function    x_out = CG_L2(y , lambda , D , varargin)

% x_out = CG_L2(y , lambda , D , varargin)
%
% The function finds the minimizer of the following quadratic regularization:
%
%               x_out = argmin_x  ||x - y||^2 + lambda ||A*x||^2
% 
% Instead of providing the function with the matrix 'A', a function handle
% "D" is used such that 
%                       D(x,1) = A*x    and     D(x,-1) = A^T*x 
%
% Other parameter listed below are optional in the sense that if they are
% not given, a default value is used:
%
% CG_L2(... , 'max_iter' , n , ...)
% "n" is the maximum number of iterations used for the conjugate gradient 
% method. The default value is length(y).
%
% CG_L2(... , 'Res_tol' , tol , ...)
% "tol" is the residual tolerance (norm of the residual vector) in the CG 
% method which is a termination condition. The default value is 1e-8.
%
% CG_L2(... , 'Anim_En' , flag , ...)
% "flag" shows if the results in the intermediate iterations are plotted or
% not. The default value is 'false'. Other inputs enable the plot.
%
% CG_L2(... , 'Orig_Sig' , x_orig , ...)
% "x_orig" is the original noiseless signal which we would like to obtain
% by this regularization method. It is only required when the animation 
% obtion is enabled.


% updating with the given inputs
[N_iter , tol , flag , x_orig]   = process_options(varargin , 'max_iter' , length(y) , ...
                'Res_tol' , 1e-8 , 'Anim_En' , 'false' , 'Orig_Sig' , []);


                            

% initial tests to check if the inputs are ok
if size(y , 1) == 1
    ys  = y.';
elseif size(y , 2) == 1
    ys  = y;
else
    error('!!! "y" should be a vector !!!')
    return
end

if size(x_orig , 1) == 1
    x_orig  = x_orig.';
else
    x_orig  = x_orig;
end

    
    
if isa(D , 'function_handle') ~= 1
    error('!!! The "D" should be a function handle !!!')
    return
end






% initialization
x_cur       = ys;   % or 0!
r_old       = ys - IpDTD(lambda , x_cur);
norm_ro     = sum(r_old.^2);
p           = r_old;
it_num      = 0;




% iterations
if strcmpi(flag , 'false')  % no animation
    
    while (it_num < N_iter)&(norm_ro > tol)
        it_num  = it_num + 1;
        
        Ap      = IpDTD(lambda , p);
        alpha   = norm_ro / (p' * Ap);
        
        x_cur   = x_cur + alpha * p;
        
        r_new   = r_old - alpha * Ap;
        norm_rn = sum(r_new.^2);
        
        p       = r_new + norm_rn/norm_ro * p;
        
        r_old   = r_new;
        norm_ro = norm_rn;
        
    end
    
    
    
else
    
    
    if ~isempty(x_orig)     % with animition and with knowing the oracle solution
        
        while (it_num < N_iter)&(norm_ro > tol)
            it_num  = it_num + 1;
            
            Ap      = IpDTD(lambda , p);
            alpha   = norm_ro / (p' * Ap);
            
            x_cur   = x_cur + alpha * p;
            
            r_new   = r_old - alpha * Ap;
            norm_rn = sum(r_new.^2);
            
            p       = r_new + norm_rn/norm_ro * p;
            
            r_old   = r_new;
            norm_ro = norm_rn;
            
            
            plot([1:length(ys)]' , x_orig ,[1:length(ys)]' , ys , '.' , [1:length(ys)]' , x_cur , '-.', 'LineWidth', 1.5)
            grid on;
            hold on;
            SNR     = 10*log10( mean(x_orig.^2) / mean( (x_orig - x_cur).^2 ));
            title(sprintf('[#iter. = %d/%d      SNR = %f dB       Residual = %f]', it_num, N_iter, SNR , norm_rn));
            legend('Main', 'Noisy', 'Est.');
            axis([-0.5 , length(ys)+0.5 , min(ys)-1 , max(ys)+1])
            hold off;
            getframe;
            
        end
        

        
    else                     % with animition but without knowing the oracle solution
        
        while (it_num < N_iter)&(norm_ro > tol)
            it_num  = it_num + 1;
            
            Ap      = IpDTD(lambda , p);
            alpha   = norm_ro / (p' * Ap);
            
            x_cur   = x_cur + alpha * p;
            
            r_new   = r_old - alpha * Ap;
            norm_rn = sum(r_new.^2);
            
            p       = r_new + norm_rn/norm_ro * p;
            
            r_old   = r_new;
            norm_ro = norm_rn;
            
            
            plot([1:length(ys)]' , ys , '.' , [1:length(ys)]' , x_cur , '-.', 'LineWidth', 1.5)
            grid on;
            hold on;
            title(sprintf('[#iter. = %d/%d      Residual = %f]', it_num, N_iter , norm_rn));
            legend('Noisy', 'Est.');
            axis([-0.5 , length(ys)+0.5 , min(ys)-1 , max(ys)+1])
            hold off;
            getframe;
            
        end
        
    end
end


% output
x_out       = reshape(x_cur , size(y));




%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Auxiliary functions called within this function
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    %--------------------%
    %----------------------------------------%
    %------------------------------------------------------------%
    function w = IpDTD(lambda , v);
        % the function implements   w =  (I + lambda* D^T*D) * v
        
        w       = v;
        for col = 1 : size(v , 2)
            w(: , col)  = w(: , col) + lambda * D( D(v(: , col) , 1) , -1);
        end
    end
    %------------------------------------------------------------%
    %----------------------------------------%
    %--------------------%



    %--------------------%
    %----------------------------------------%
    %------------------------------------------------------------%
    % PROCESS_OPTIONS - Processes options passed to a Matlab function.
    %                   This function provides a simple means of
    %                   parsing attribute-value options.  Each option is
    %                   named by a unique string and is given a default
    %                   value.
    % Copyright (C) 2002 Mark A. Paskin
    function [varargout] = process_options(args, varargin)
        
        % Check the number of input arguments
        n = length(varargin);
        if (mod(n, 2))
            error('Each option must be a string/value pair.');
        end
        
        % Check the number of supplied output arguments
        if (nargout < (n / 2))
            error('Insufficient number of output arguments given');
        elseif (nargout == (n / 2))
            warn = 1;
            nout = n / 2;
        else
            warn = 0;
            nout = n / 2 + 1;
        end
        
        % Set outputs to be defaults
        varargout = cell(1, nout);
        for i=2:2:n
            varargout{i/2} = varargin{i};
        end
        
        % Now process all arguments
        nunused = 0;
        for i=1:2:length(args)
            found = 0;
            for j=1:2:n
                if strcmpi(args{i}, varargin{j})
                    varargout{(j + 1)/2} = args{i + 1};
                    found = 1;
                    break;
                end
            end
            if (~found)
                if (warn)
                    warning(sprintf('Option ''%s'' not used.', args{i}));
                    args{i}
                else
                    nunused = nunused + 1;
                    unused{2 * nunused - 1} = args{i};
                    if length(args)>i
                        unused{2 * nunused} = args{i + 1};
                    end
                end
            end
        end
        
        % Assign the unused arguments
        if (~warn)
            if (nunused)
                varargout{nout} = unused;
            else
                varargout{nout} = cell(0);
            end
        end
    end
    %------------------------------------------------------------%
    %----------------------------------------%
    %--------------------%





    
end