function    x_out = Duality_L1(y , lambda , D , varargin)

% x_out = Duality_L1(y , lambda , D ,  L)
%
% The function finds the minimizer of the following l1-norm regularization:
%
%               x_out = argmin_x  ||x - y||^2 + lambda ||A*x||_1
% 
% Instead of providing the function with the matrix 'A', a function handle
% "D" is used such that 
%                       D(x,1) = A*x    and     D(x,-1) = A^T*x 
%
% Other parameter listed below are optional in the sense that if they are
% not given, a default value is used:
%
% Duality_L1(... , 'max_iter' , n , ...)
% "n" is the maximum number of iterations used for the iterative methods. 
% The default value is 100.
%
% Duality_L1(... , 'Lipschitz' , l , ...)
% "l" is the Lipshitz constant of the matrix A (maximum eigenvalue of A*A^T).
% If this is not given, the function finds it, but it increases the 
% execution time.
%
% Duality_L1(... , 'DGap_tol' , tol , ...)
% "tol" is the tolerance for the Duality Gap which is a termination
% condition. The default value is 1e-5.
%
% Duality_L1(... , 'Anim_En' , flag , ...)
% "flag" shows if the results in the intermediate iterations are plotted or
% not. The default value is 'false'. Other inputs enable the plot.
%
% Duality_L1(... , 'Orig_Sig' , x_orig , ...)
% "x_orig" is the original noiseless signal which we would like to obtain
% by this L1 method. It is only required when the animation obtion is
% enabled.


% updating with the given inputs
[N_iter , L , tol , flag , x_orig]   = process_options(varargin , 'max_iter' , 100 , ...
                'Lipschitz' , [] , 'DGap_tol' , 1e-5 , 'Anim_En' , 'false' , 'Orig_Sig' , []);


                            

% initial tests to check if the inputs are ok
if size(y , 1) == 1
    ys  = y.';
elseif size(y , 2) == 1
    ys  = y;
else
    error('!!! "y" should be a vector !!!')
    return
end

if size(x_orig , 1) == 1
    x_orig  = x_orig.';
else
    x_orig  = x_orig;
end

    
    
if isa(D , 'function_handle') ~= 1
    error('!!! The "D" should be a function handle !!!')
    return
end



% Primal and Dual cost functions
Prim_cost   = @(x) norm(x-ys)^2 + lambda* norm(D(x , 1) , 1);
Dual_cost   = @(v) -lambda^2/4 * v' * D(D(v , -1) , 1) + lambda * ys' * D(v , -1);




% initialization
it_num      = 0;
Duality_Gap = tol + 1;
v_Nestr     = sign( D(ys,1) )  /  2;
x_Nestr     = v_Nestr;
y_Nestr     = v_Nestr;

x_Prim      = ys - 0.5 * lambda * D(x_Nestr , -1);

theta_o     = 1;
theta_n     = 2/3;
ff          = -lambda * D(ys,1).';

if isempty(L)
    t       = eigs( @(x)DDT(lambda,x) ,  length(v_Nestr)  , 1 ) ^ (-1);
else
    t       = 2 / (lambda^2 * L);
end




% iterations
if strcmpi(flag , 'false')      % no animation
    
    while (it_num < N_iter)&(Duality_Gap > tol)
        it_num  = it_num + 1;
        
        v_Nestr     = v_Nestr - t / theta_o * (DDT(lambda , y_Nestr) + ff');
        v_Nestr(v_Nestr >  1)   =  1;
        v_Nestr(v_Nestr < -1)   = -1;
        
        x_Nestr     = (1 - theta_o) * x_Nestr + theta_o * v_Nestr;
        
        y_Nestr     = (1 - theta_n) * x_Nestr + theta_n * v_Nestr;
        
        theta_o     = theta_n;
        theta_n     = 2/(it_num+2);
        
        % Duality Gap
        x_Prim      = ys - 0.5 * lambda * D(x_Nestr , -1);
        Duality_Gap = Prim_cost(x_Prim) - Dual_cost(x_Nestr);
        
    end
    
    
else
    
    
    if ~isempty(x_orig)     % with animition and with knowing the oracle solution
        
        while (it_num < N_iter)&(Duality_Gap > tol)
            it_num  = it_num + 1;
            
            v_Nestr     = v_Nestr - t / theta_o * (DDT(lambda , y_Nestr) + ff');
            v_Nestr(v_Nestr >  1)   =  1;
            v_Nestr(v_Nestr < -1)   = -1;
            
            x_Nestr     = (1 - theta_o) * x_Nestr + theta_o * v_Nestr;
            
            y_Nestr     = (1 - theta_n) * x_Nestr + theta_n * v_Nestr;
            
            theta_o     = theta_n;
            theta_n     = 2/(it_num+2);
            
            x_Prim      = ys - 0.5 * lambda * D(x_Nestr , -1);
            
            Duality_Gap     = Prim_cost(x_Prim) - Dual_cost(x_Nestr);
            SNR             = 10*log10( mean(x_orig.^2) / mean( (x_orig - x_Prim).^2 ));
            
            plot([1:length(ys)]' , x_orig ,[1:length(ys)]' , ys , '.' , [1:length(ys)]' , x_Prim , '-.', 'LineWidth', 1.5)
            grid on;
            hold on;
            title(sprintf('[#iter. = %d/%d      SNR = %f dB       Duality-Gap = %f]', it_num, N_iter, SNR , Duality_Gap));
            legend('Main', 'Noisy', 'Est.');
            axis([-0.5 , length(ys)+0.5 , min(ys)-1 , max(ys)+1])
            hold off;
            getframe;
            
        end
        
    else                    % with animition but without knowing the oracle solution
         
        while (it_num < N_iter)&(Duality_Gap > tol)
            it_num  = it_num + 1;
            
            v_Nestr     = v_Nestr - t / theta_o * (DDT(lambda , y_Nestr) + ff');
            v_Nestr(v_Nestr >  1)   =  1;
            v_Nestr(v_Nestr < -1)   = -1;
            
            x_Nestr     = (1 - theta_o) * x_Nestr + theta_o * v_Nestr;
            
            y_Nestr     = (1 - theta_n) * x_Nestr + theta_n * v_Nestr;
            
            theta_o     = theta_n;
            theta_n     = 2/(it_num+2);
            
            x_Prim      = ys - 0.5 * lambda * D(x_Nestr , -1);
            
            Duality_Gap     = Prim_cost(x_Prim) - Dual_cost(x_Nestr);
            
            plot([1:length(ys)]' , ys , '.r' , [1:length(ys)]' , x_Prim , '-.', 'LineWidth', 1.5)
            grid on;
            hold on;
            title(sprintf('[#iter. = %d/%d       Duality-Gap = %f]', it_num, N_iter , Duality_Gap));
            legend('Noisy', 'Est.');
            axis([-0.5 , length(ys)+0.5 , min(ys)-1 , max(ys)+1])
            hold off;
            getframe;
            
        end
        
    end

end


% output
x_out       = reshape(x_Prim , size(y));




%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Auxiliary functions called within this function
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    %--------------------%
    %----------------------------------------%
    %------------------------------------------------------------%
    function w = DDT(lambda , v);
        % the function implements lambda^2/2 * D*D^T*v
        
        w       = zeros(size(v));
        for col = 1 : size(v , 2)
            w(: , col)  = lambda^2 / 2 * D( D(v(: , col) , -1) , 1);
        end
    end
    %------------------------------------------------------------%
    %----------------------------------------%
    %--------------------%
    
    
    %--------------------%
    %----------------------------------------%
    %------------------------------------------------------------%
    function w = DzDzT(v);
        
        w       = zeros(size(v));
        for col = 1 : size(v , 2)
            alaki       = zeros(m , 1);
            alaki(zero_ind)     = v(: , col);
            alaki       = D( D(alaki , -1) , 1);
            w(: , col)  = alaki(zero_ind);
        end
    end
    %------------------------------------------------------------%
    %----------------------------------------%
    %--------------------%



    %--------------------%
    %----------------------------------------%
    %------------------------------------------------------------%
    % PROCESS_OPTIONS - Processes options passed to a Matlab function.
    %                   This function provides a simple means of
    %                   parsing attribute-value options.  Each option is
    %                   named by a unique string and is given a default
    %                   value.
    % Copyright (C) 2002 Mark A. Paskin
    function [varargout] = process_options(args, varargin)
        
        % Check the number of input arguments
        n = length(varargin);
        if (mod(n, 2))
            error('Each option must be a string/value pair.');
        end
        
        % Check the number of supplied output arguments
        if (nargout < (n / 2))
            error('Insufficient number of output arguments given');
        elseif (nargout == (n / 2))
            warn = 1;
            nout = n / 2;
        else
            warn = 0;
            nout = n / 2 + 1;
        end
        
        % Set outputs to be defaults
        varargout = cell(1, nout);
        for i=2:2:n
            varargout{i/2} = varargin{i};
        end
        
        % Now process all arguments
        nunused = 0;
        for i=1:2:length(args)
            found = 0;
            for j=1:2:n
                if strcmpi(args{i}, varargin{j})
                    varargout{(j + 1)/2} = args{i + 1};
                    found = 1;
                    break;
                end
            end
            if (~found)
                if (warn)
                    warning(sprintf('Option ''%s'' not used.', args{i}));
                    args{i}
                else
                    nunused = nunused + 1;
                    unused{2 * nunused - 1} = args{i};
                    if length(args)>i
                        unused{2 * nunused} = args{i + 1};
                    end
                end
            end
        end
        
        % Assign the unused arguments
        if (~warn)
            if (nunused)
                varargout{nout} = unused;
            else
                varargout{nout} = cell(0);
            end
        end
    end
    %------------------------------------------------------------%
    %----------------------------------------%
    %--------------------%





    
end