function xhat = messagePassingTime(y, sgrid, prior, vn, verbose, x)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Set default parameters
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if(~exist('verbose', 'var'))
    verbose = 0;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Algorithm body
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Signal length
n       = length(y);

% Number of iterations
T       = n - 1;

% Initialize messages
[cl, cr]    = initialize(prior, sgrid, n);

% Initialize estimate
xhat    = y;

% Prior probability
p       = priordist(sgrid, prior);

% Print initial messages
if (verbose == 1)
    fprintf('start: messagePassingTime\n');
elseif(verbose == 2)
    %figure('Color', 'w');
end

% Perform iterations
for t = 1:T
    % Left side update
    phi         = gaussian(sgrid, y(t), vn);
    vl          = normalize(phi .* cl(t, :));
    cl(t+1, :)  = normalize(convolve(p, vl));
    
    % Right side update
    phi         = gaussian(sgrid, y(n+1-t), vn);
    vr          = normalize(phi .* cr(n+1-t, :));
    cr(n-t, :)  = normalize(convolve(p, vr));
    
    % Perform left estimation
    apostl      = normalize(cl(t+1, :) .* cr(t+1, :) .* gaussian(sgrid, y(t+1), vn));
    xhat(t+1)   = sum(sgrid .* apostl);
    
    % Perform right estimation
    apostr      = normalize(cl(n-t, :) .* cr(n-t, :) .* gaussian(sgrid, y(n-t), vn));
    xhat(n-t)   = sum(sgrid .* apostr);
    
    % Print to command line
    mse         = 10 * log10( sum((x-xhat).^2) / n );
    if (verbose == 1 && exist('x', 'var'))
        
        
        st      = sprintf('[%%%d.d/%%d] MSE: %%.4f', 1+floor(log10(T)));
        fprintf(st, t, T, mse);
        fprintf('\n');
        
    elseif (verbose == 2 && exist('x', 'var'))
        
        plot(1:n, x, 1:n, y, '.', 1:n, xhat, '-.', 'LineWidth', 1.5);
        grid on;
        hold on;
        plot([t+1-0.5 t+1-0.5], [min(y), max(y)], 'g');
        plot([t+1+0.5 t+1+0.5], [min(y), max(y)], 'g');
        plot([n-t-0.5 n-t-0.5], [min(y), max(y)], 'c');
        plot([n-t+0.5 n-t+0.5], [min(y), max(y)], 'c');
        title(sprintf('[t = %d/%d: MSE = %f]', t, T, mse));
        legend('Main', 'Noisy', 'Est.');
        axis([-0.5 , n+0.5 , min(y)-1 , max(y)+1])
        hold off;

        getframe;
    end
    
end

% Print final message
if(verbose == 1)
    fprintf('end: messagePassingTime\n');
end








%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Helper functions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [cl, cr]  = initialize(prior, sgrid, n)
% INITIALIZE messages of the algorithm.

% Number of samples
N       = length(sgrid);

% Left constraints
cl      = ones(n, N);
cl(1, :) = priordist(sgrid, prior);

% Right constraints
cr      = ones(n, N);



%--- --- ---  --  --  --   -   -   -   -    -    -     -     -      -



function p = priordist(sgrid, prior)
% PRIORDIST generates sampled prior distribution

switch(prior.type)
    case{'g', 'gaussian'}
        p = gaussian(sgrid, 0, prior.param);
    case{'l', 'laplace'}
        p = laplace(sgrid, prior.param);
    case{'s', 'student'}
        p = student(sgrid, prior.param);
    case{'a', 'alpha_stable'}
        p = alpha(sgrid, 1 , prior.param);

    case{'pg', 'poisson-gaussian'}
        p = poissonMix('g', sgrid, prior);
    case{'pl', 'poisson-laplace'}
        p = poissonMix('l', sgrid, prior);
    case{'ps', 'poisson-student'}
        p = poissonMix('s', sgrid, prior);
    case{'pa', 'poisson-alpha_stable'}
        p = poissonMix('a', sgrid, prior);
        
    otherwise
        error('messagePassingTime: Prior not supported!');
end



%--- --- ---  --  --  --   -   -   -   -    -    -     -     -      -



function phi = poissonMix(type, sgrid, prior)
% poisson-gaussian or ...

% Sampling interval
delta   = abs(  sgrid(2) - sgrid(1)  );
OSR     = 10;
sgridF  = linspace( sgrid(1)-delta/2  ,  sgrid(end)+delta/2  ,  OSR * length(sgrid)+1 );
deltaF  = abs( sgridF(2) - sgridF(1) );


% Indices
i       = sgrid / delta;

% Kronecker delta
kdelta  = (i==0);

% Non-dirac distribution
switch(type)
    case{'g', 'gaussian'}
        p = gaussian(sgridF , 0 , prior.param);
    case{'l', 'laplace'}
        p = laplace(sgridF ,  prior.param);
    case{'s', 'student'}
        p = student(sgridF ,  prior.param);
    case{'a', 'alpha_stable'}
        p = alpha(sgridF , 1 , prior.param);
    otherwise
        error('poissonMix: Prior not supported!');
end

% Non-dirac part
p       = p / deltaF;
lambda  = -log( 1 - prior.rho );
Nterms  = 20;
pn      = [1/deltaF];
coef    = 1;
phi     = 0;
for nn = 1 : Nterms
    pn  = conv(p , pn , 'same') * deltaF;
%     sum(pn)*deltaF
    coef= coef * lambda / nn;
    phi = phi + coef * pn;
end

% sampling and adding the dirac part
phi     = (  sum( reshape(phi(1:end-1) , OSR , length(sgrid)) ) + sum( reshape(phi(2:end) , OSR , length(sgrid)) )  )  * deltaF/2;
phi     = (1-prior.rho) * (kdelta + phi);
% phi     = phi / sum(phi);



% %--- --- ---  --  --  --   -   -   -   -    -    -     -     -      -
% 
% 
% 
% function phi = cauchy(sgrid, s)
% % CAUCHY generates sampled cauchy distribution
% 
% % Sample from Laplace
% phi     = s ./ ( pi * ( (s^2) + (sgrid.^2) ) );
% phi     = phi/sum(phi);



%--- --- ---  --  --  --   -   -   -   -    -    -     -     -      -



function phi = gaussian(sgrid, mu, va)
% GAUSSIAN generates sampled gaussian distribution

% Sampling interval
delta   = abs(sgrid(2) - sgrid(1));

% Indices
i       = sgrid / delta;

% Standard deviation
s       = sqrt(va);

% Sample from Gaussian
phi     = normcdf(delta * (i+0.5)  ,  mu  ,  s) - normcdf(delta * (i-0.5) , mu , s);
phi     = phi / sum(phi);



%--- --- ---  --  --  --   -   -   -   -    -    -     -     -      -



function phi = laplace(sgrid, lambda)
% LAPLACE generates sampled laplace distribution

% Sample from Laplace
phi     = 0.5 * lambda*exp(-lambda * abs(sgrid));
phi     = phi / sum(phi);



%--- --- ---  --  --  --   -   -   -   -    -    -     -     -      -



function phi = student(sgrid, q)
% STUDENT generates sampled student's-t distribution

% Sampling interval
delta   = abs(sgrid(2) - sgrid(1));

% Indices
i       = sgrid / delta;

% Sample from Student's-t
phi     = cdf('t' , delta * (i+0.5) , q) - cdf('t' , delta * (i-0.5) , q);
phi     = phi / sum(phi);



%--- --- ---  --  --  --   -   -   -   -    -    -     -     -      -



function phi = alpha(sgrid, sigma , alpha)
% ALPHA generates sampled alpha-stable distribution

% Sampling interval
delta   = abs(sgrid(2) - sgrid(1));
i       = sgrid / delta;    % Indices
    
if alpha == 1   % cauchy distribution
    
    % Sample from Student's-t
    phi     = 1/pi  *  ( atan( delta * (i+0.5) / sigma ) - atan( delta * (i-0.5) / sigma ) ) ;
    phi     = phi / sum(phi);

else
    
    phi     = zeros(1 , length(sgrid));
    
    % Compute CDF
    V       = @(theta)    ( cos(theta) ./ sin( alpha*theta ) ) .^ (alpha / (alpha-1)) ...
                       .* cos( (alpha - 1) * theta ) ./ cos(theta);
    
    sgridP  = sgrid(sgrid > 0);
    gshiftU = ( (sgridP + 0.5*delta) / sigma ) .^ ( alpha / (alpha-1) );
    gshiftL = ( (sgridP - 0.5*delta) / sigma ) .^ ( alpha / (alpha-1) );
    Ppart   = sign(1-alpha)/pi * ...
              ( quadv(@(theta) exp(-gshiftU * V(theta)) , 1e-10 , pi/2-1e-10 , 1e-8) - ...
                quadv(@(theta) exp(-gshiftL * V(theta)) , 1e-10 , pi/2-1e-10 , 1e-8) );
                   
                   
    phi(sgrid > 0)  = Ppart;
    phi(sgrid < 0)  = Ppart(end:-1:1);
    phi(sgrid == 0) = 2 * ( (alpha > 1) + (alpha < 1) * 0.5 + sign(1-alpha)/pi * ...
                     ( quadv(@(theta) exp(- (0.5*delta/sigma) .^ ( alpha / (alpha-1) ) * V(theta)) , 1e-10 , pi/2-1e-10 , 1e-8)) ) ...
                      - 1;
    
end
    
phi     = phi / sum(phi);



%--- --- ---  --  --  --   -   -   -   -    -    -     -     -      -



function pout = normalize(pin)
pout    = pin / sum(pin);




%--- --- ---  --  --  --   -   -   -   -    -    -     -     -      -




function f = convolve(x, h)
% CONVOLVE fast centered convolution using fftfilt.

% Number of samples
N       = length(x);

% Check if N is odd
assert(mod(N, 2) == 1, 'convolve: PDFs must have even number of samples.');

% Pad-input vectos to have 2N-1 elements
x       = [x, zeros(1, N-1)];
h       = [h, zeros(1, N-1)];

% Fast convolution
f       = fftfilt(x, h);

% Indeces to trim off
fbeg    = (N+1)/2;
fend    = (3*N-1)/2;

% Trim
f       = f(fbeg:fend);