function [SS_data , SS_data_N] = priorGen(handles)

% [SS_data ; SS_data_N] = priorGen(handles)
%
% The function generates samples of the generalized sparse signals with the
% parameter structure "handles":
%
%   'handles.Prior_Type':   1 => Simple process (no mass)
%                           2 => Compound Poisson (with mass)
%
%   'handles.MassProb':     the value of the mass (Compound Poisson)
%
%   'handles.Dist_Sel':     1 => zero-mean Gaussian Distribution
%                           2 => zero-mean Laplace Distribution
%                           3 => zero-mean Student's-t Distribution
%                           4 => symmetric alpha_stable with sigma = 1
%
%   'handles.Dist_Param':   Gaussian -> Variance  [exp(-x^2 / (2*var^2)) / (2*pi*var^2)^5]
%                           Laplace  -> b         [exp(-|x| / b) / (2*b)]
%                           Student  -> q         [(1 + x^2/q) ^ (-(q+1)/2 * const.]
%                           alpha_st -> alpha
%
%   'handles.SS_length':    Length of the output discrete signal
%
%   'handles.Noise_Var':    Variance of the AWGN
%
%   'handles.Operator':     Order of Integration (e.g., 1 for Levy)
%
%
%   "SS_data":      The output discrete signal without AWGN
%   "SS_data_N":    The output discrete signal with AWGN


% generating a white noise
switch handles.Prior_Type
    
    case 1      % simple process without mass
        dx      = 1 / 50;
        x       = [0 : dx : handles.SS_length];
        
        % white noise generation
        switch handles.Dist_Sel
            case 1      % normal
                whitenoise      = random('norm' , 0 , handles.Dist_Param*sqrt(dx) , 1 , length(x));
                
                
            case 2      % Laplace
                rand_aux        = random('unif' , -0.5 , 0.5 , 1 , handles.SS_length);
                whitenoise      = -handles.Dist_Param * sign(rand_aux) .* log(1 - 2 * abs(rand_aux));
                
                
            case 3      % Student's-t
                normal_RV       = randn(1 , handles.SS_length);
                XiSquare_RV     = random('chi2' , handles.Dist_Param , 1 , handles.SS_length);
                whitenoise      = sqrt(handles.Dist_Param) * normal_RV ./ (XiSquare_RV .^ 0.5);
                
                
            case 4      % alpha_stable
                alpha       = handles.Dist_Param;
                theta       = random('uniform' , -pi/2 , pi/2 , 1 , length(x));
                w           = random('exp' , 1 , 1 , length(x));
                whitenoise  =    sin(alpha * theta) ./ (cos(theta)) .^ (1/alpha) ...
                              .* (cos((1-alpha) * theta) ./ w) .^ ((1-alpha) / alpha);
                whitenoise  = whitenoise * (dx) ^ (1 / alpha);
                
                
            otherwise
                error('Error:    Unrecognizable probability distribution')
                
        end
        
        
        
        % Integration operator
        SS_data         = whitenoise;
        for Int_ind = 1 : handles.Operator
            SS_data     = cumsum(SS_data);
        end
        step            = round(length(SS_data) / handles.SS_length);
        SS_data         = SS_data(1 : step : handles.SS_length*step);
        
        
        
        
    case 2      % Compound Poisson
        lambda  = -log(handles.MassProb);
        dx      = 1 / 50;
        x       = [0 : dx : handles.SS_length];
        
        % impulse-train Poisson process
        y_imp           = random('uniform' , 0 , 1 , size(x));
        prob            = 1 - lambda * dx;
        y_imp(y_imp >  prob)    = 1;
        y_imp(y_imp <= prob)    = 0;
        x_imp           = x(y_imp == 1);
        x_imp           = union([0] , x_imp);
        N_jump          = length(x_imp);
        
        
        % white noise generation
        switch handles.Dist_Sel
            case 1      % normal
                JumpValues      = random('norm' , 0 , handles.Dist_Param , 1 , N_jump);
                
                
            case 2      % Laplace
                rand_aux        = random('unif' , -0.5 , 0.5 , 1 , N_jump);
                JumpValues      = -handles.Dist_Param * sign(rand_aux) .* log(1 - 2 * abs(rand_aux));
                
                
            case 3      % Student's-t
                normal_RV       = randn(1 , N_jump);
                XiSquare_RV     = random('chi2' , handles.Dist_Param , 1 , N_jump);
                JumpValues      = sqrt(handles.Dist_Param) * normal_RV ./ (XiSquare_RV .^ 0.5);
                
                
            case 4      % alpha_stable
                alphaV      = handles.Dist_Param;
                theta       = random('uniform' , -pi/2 , pi/2 , 1 , N_jump);
                w           = random('exp' , 1 , 1 , N_jump);
                JumpValues  =    sin(alphaV * theta) ./ (cos(theta)) .^ (1/alphaV) ...
                              .* (cos((1-alphaV) * theta) ./ w) .^ ((1-alphaV) / alphaV);
                
            otherwise
                error('Error:    Unrecognizable probability distribution')
                
        end
        
        
        % Integration operator
        SS_data         = 0 * x;
        p0              = random('uniform' , 0 , 1 , 1 , 1);
        JumpValues(1)   = JumpValues(1) * (p0 > handles.MassProb);
        for imp_ind = 1 : length(x_imp)
            [vv , ii]   = min( abs(x_imp(imp_ind) - x) );
            SS_data(ii) = JumpValues(imp_ind) / dx;
        end
        
        for aa = 1 : handles.Operator
            SS_data         = cumsum(SS_data) * dx;
        end
        step            = round(1 / dx);
        SS_data         = downsample(SS_data , step);
        SS_data         = SS_data(length(SS_data)-handles.SS_length+1 : end);
        
        
end





% adding noise
noise       = random('normal' , 0 , sqrt(handles.Noise_Var) , 1 , handles.SS_length);
SS_data_N   = SS_data + noise;